#: Forecast types requestable through nemseer.
#: See also :term:`forecast types`, and :term:`pre-dispatch` and :term:`PASA`.
FORECAST_TYPES = ("P5MIN", "PREDISPATCH", "PDPASA", "STPASA", "MTPASA")

MMSDM_ARCHIVE_URL = "http://www.nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM/"
"""Wholesale electricity data archive base URL"""

#: Tables which should be directed to the PREDISP_ALL_DATA URL.
#: The corresponding tables in the DATA folder (which end with "_D") only contain the
#: latest forecasted value
PREDISP_ALL_DATA = ("CONSTRAINT", "INTERCONNECTORRES", "PRICE", "LOAD", "REGIONSUM")

MTPASA_DUID_URL = "http://nemweb.com.au/Reports/Current/MTPASA_DUIDAvailability/"
"""MTPASA DUID Availability"""

#: Enumerated tables for each forecast type
#: First element of tuple is table name
#: Second element of tuple is number which to enumerate table to
ENUMERATED_TABLES = {
    "P5MIN": [("CONSTRAINTSOLUTION", 4)],
    "PREDISPATCH": [("CONSTRAINT", 2), ("LOAD", 2)],
}

DEPRECATED_TABLES = {
    "MTPASA": [
        "CASESOLUTION",
    ],
}
"""
Deprecated tables
"""

DATETIME_FORMAT = "%Y/%m/%d %H:%M"
"""
`nemseer` date format
"""

RUNTIME_COL = {
    "P5MIN": "RUN_DATETIME",
    "PREDISPATCH": "PREDISPATCH_RUN_DATETIME",
    "PDPASA": "RUN_DATETIME",
    "STPASA": "RUN_DATETIME",
    "MTPASA": "RUN_DATETIME",
}
"""
If it exists, `nemseer` will use the corresponding column for `run` time filtering.
"""

FORECASTED_COL = {
    "P5MIN": "INTERVAL_DATETIME",
    "PREDISPATCH": "DATETIME",
    "PDPASA": "INTERVAL_DATETIME",
    "STPASA": "INTERVAL_DATETIME",
    "MTPASA": "DAY",
}
"""
If it exists, `nemseer` uses the corresponding column for `forecasted` time filtering.
"""

ID_COLS = {
    "STUDYREGIONID",
    "CONSTRAINTID",
    "INTERCONNECTORID",
    "DUID",
    "CONNECTIONPOINTID",
    "PARTICIPANTID",
    "EXPORTGENCONID",
    "IMPORTGENCONID",
    "REGIONID",
    "LINKID",
    "USE_ITERATION_ID",
    "RESERVELIMITID",
    "SCENARIO",
}

TYPE_COLS = {
    "BIDTYPE",
    "RUNTYPE",
    "RUN_NO",
    "DEMAND_POE_TYPE",
    "AGGREGATION_PERIOD",
    "PERIOD_ENDING",
    "EFFECTIVEDATE",
    "VERSION_DATETIME",
}

DATETIME_COLS = {
    "DATETIME",
    "EFFECTIVEDATE",
    "INTERVAL_DATETIME",
    "RUN_DATETIME",
    "AUTHORISEDDATE",
    "LASTCHANGED",
    "VERSION_DATETIME",
    "DAY",
    "PUBLISH_DATETIME",
    "LATEST_OFFER_DATETIME",
    "STARTDATE",
    "ENDDATE",
    "PERIOD_ENDING",
    "GENCONID_EFFECTIVEDATE",
    "BIDSETTLEMENTDATE",
    "SETTLEMENTDATE",
    "OFFERDATE",
}

INVALID_STUBS_FILE = ".invalid_aemo_files.txt"
"""File in :term:`raw_cache` that contains invalid/corrupted AEMO files"""

USER_AGENTS = [
    (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
        + "(KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36)"
    ),
    "Mozilla/5.0 (Windows NT 5.1; rv:7.0.1) Gecko/20100101 Firefox/7.0.1",
    (
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 "
        + "(KHTML, like Gecko) Chrome/44.0.2403.157 Safari/537.36"
    ),
    (
        "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:49.0) "
        + "Gecko/20100101 Firefox/49.0"
    ),
    (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) "
        + "AppleWebKit/605.1.15 (KHTML, like Gecko) Version/"
        + "11.1.2 Safari/605.1.15"
    ),
    (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/601.7.7 "
        + "(KHTML, like Gecko) Version/9.1.2 Safari/601.7.7"
    ),
    (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.11; rv:45.0) "
        + "Gecko/20100101 Firefox/45.0"
    ),
    (
        "Mozilla/5.0 (X11; CrOS x86_64 14268.67.0) AppleWebKit/537.36 "
        + "(KHTML, like Gecko) Chrome/96.0.4664.111 Safari/537.36)"
    ),
    (
        "Opera/9.80 (Linux armv7l) Presto/2.12.407 Version/12.51 , "
        + "D50u-D1-UHD/V1.5.16-UHD (Vizio, D50u-D1, Wireless))"
    ),
    "Wget/1.12 (linux-gnu)",
    "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",
    "Mozilla/5.0 (Windows NT 6.1; Trident/7.0; rv:11.0) like Gecko",
    (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
        + "(KHTML, like Gecko) Chrome/70.0.3538.102 Safari/537.36 Edge/18.18363"
    ),
    (
        "Mozilla/5.0 (X11; Linux x86_64; U; en-us) AppleWebKit/537.36 (KHTML, "
        + "like Gecko) Silk/3.68 like Chrome/39.0.2171.93 Safari/E7FBAF)"
    ),
    (
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/534.34 (KHTML, like Gecko) "
        + "Qt/4.8.1 Safari/E7FBAF"
    ),
    (
        "Mozilla/5.0 (X11; Linux armv7l) AppleWebKit/537.36 (KHTML, like Gecko)"
        + "Raspbian Chromium/74.0.3729.157 Chrome/74.0.3729.157 Safari/537.36"
    ),
]



# read version from installed package
import logging
import sys
import warnings
from importlib.metadata import version

# aliases for SQLLoader functions
from .data import FORECAST_TYPES as forecast_types
from .downloader import get_sqlloader_forecast_tables as get_tables
from .downloader import get_sqlloader_years_and_months as get_data_daterange
from .forecast_type.run_time_generators import generate_runtimes
from .nemseer import compile_data, download_raw_data

__version__ = version("nemseer")

logging.getLogger(__name__).addHandler(logging.NullHandler())
logging.basicConfig(
    stream=sys.stdout, level=logging.INFO, format="%(levelname)s: %(message)s"
)

warnings.simplefilter(action="ignore", category=FutureWarning)

__all__ = [
    "compile_data",
    "download_raw_data",
    "get_tables",
    "get_data_daterange",
    "generate_runtimes",
    "forecast_types",
]



import logging
from datetime import datetime
from pathlib import Path
from typing import List, Tuple, Union

import pandas as pd
import xarray as xr

from .data import (
    DATETIME_COLS,
    DATETIME_FORMAT,
    FORECASTED_COL,
    ID_COLS,
    RUNTIME_COL,
    TYPE_COLS,
)

logger = logging.getLogger(__name__)


def _parse_datetime_cols(df: pd.DataFrame) -> pd.DataFrame:
    """Finds datetime columns in the DataFrame and converts them to datetime

    Args:
        df: pandas.DataFrame
    Returns:
        :class:`pandas.DataFrame` with datetime columns converted according to standard
        AEMO
        format
    """
    dt_cols = DATETIME_COLS
    dt_cols_present = dt_cols.intersection(set(df.columns.tolist()))
    for col in dt_cols_present:
        df.loc[:, col] = pd.to_datetime(df[col], format=DATETIME_FORMAT)
    return df


def _parse_id_cols(df: pd.DataFrame) -> pd.DataFrame:
    """Finds relevant ID columns in the DataFrame and converts them to categories

    Args:
        df: pandas.DataFrame
    Returns:
        :class:`pandas.DataFrame` with ID columns converted to categories
    """
    id_cols = {
        "CONSTRAINTID",
        "INTERCONNECTORID",
        "DUID",
        "CONNECTIONPOINTID",
        "PARTICIPANTID",
        "EXPORTGENCONID",
        "IMPORTGENCONID",
        "REGIONID",
    }
    id_cols_present = id_cols.intersection(set(df.columns.tolist()))
    for col in id_cols_present:
        df.loc[:, col] = df[col].astype("category")
    return df


def _parse_predispatch_seq_no(df: pd.DataFrame) -> pd.DataFrame:
    """Parses `PREDISPATCHSEQNO` as datetime and adds `PREDISPATCH_RUN_DATETIME`

    Args:
        df: pandas.DataFrame
    Returns:
        :class:`pandas.DataFrame` with additional column `PREDISPATCH_RUN_DATETIME`
    """
    df["PREDISPATCHSEQNO"] = df["PREDISPATCHSEQNO"].astype(int).astype(str)
    parsed = df["PREDISPATCHSEQNO"].str.extract(r"^([0-9]{8})([0-9]{2})$")
    year_month_day = pd.to_datetime(parsed[0], format="%Y%m%d")
    hour_min = ((parsed[1].astype(int) - 1) * pd.Timedelta(minutes=30)).add(
        pd.Timedelta(hours=4, minutes=30)
    )
    df["PREDISPATCH_RUN_DATETIME"] = year_month_day + hour_min  # type: ignore
    return df


def clean_forecast_csv(filepath_or_buffer: Union[str, Path]) -> pd.DataFrame:
    """Given a forecast csv filepath or buffer, reads and cleans the forecast csv.

    Cleans artefacts in the forecast csv files, including AEMO metadata at start of file
    and end of report line. Also removes any duplicate rows.

    Args:
        filepath_or_buffer: As for :func:`pandas.read_csv`
    Returns:
        Cleaned :class:`pandas.DataFrame` with forecast data
    Warning:
        Removes duplicate rows. Raises a warning when doing so.
    """
    df = pd.read_csv(filepath_or_buffer, skiprows=1, low_memory=False)
    # remove end of report line
    df = df.iloc[0:-1, :]
    # skip AEMO metadata
    drop_cols = df.columns.tolist()[0:4]
    df = df.drop(drop_cols, axis="columns")
    df = _parse_datetime_cols(df)
    df = _parse_id_cols(df)
    if "PREDISPATCHSEQNO" in df.columns:
        df = _parse_predispatch_seq_no(df)
    for col in [col for col in df.columns if df.dtypes[col] == "float64"]:
        df[col] = pd.to_numeric(df[col], downcast="integer")
    for col in [col for col in df.columns if df.dtypes[col] == "float64"]:
        df[col] = pd.to_numeric(df[col], downcast="float")
    if any(dup_df := df.duplicated()):
        dup_rows = dup_df.loc[dup_df is True]
        logger.warning(
            "Duplicate rows detected. Dropping the following rows:\n" + f"{dup_rows}"
        )
        df = df.drop_duplicates()
    return df


def _filter_on_datetime_col(
    df: pd.DataFrame, dt_col: str, start: datetime, end: datetime
) -> pd.DataFrame:
    """Filter the given datetime column based on the supplied start and end times

    Args:
        df: pandas.DataFrame
        dt_col: Datetime columns in :attr:`df`
        start: Start datetime
        end: End datetime
    Returns
        DataFrame with datetime filtering applied on :attr:`dt_col`.
    """
    df = df.loc[df[dt_col] >= start, :]
    df = df.loc[df[dt_col] <= end, :]
    return df


def apply_run_and_forecasted_time_filters(
    df: pd.DataFrame,
    run_start: datetime,
    run_end: datetime,
    forecasted_start: datetime,
    forecasted_end: datetime,
    forecast_type: str,
) -> pd.DataFrame:
    """Applies filtering for run times (i.e. :term:`run_start` and :term:`run_end`) and
    forecasted times (i.e. :term:`forecasted_start` and :term:`forecasted_end`).

    Datetime filtering is applied to a column fetched from lookup tables that map
    the relevant column name to each :term:`forecast type`. If the run time/forecasted
    column obtained from the lookup is not present in the DataFrame, the respective
    filter is not applied.

    Args:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
        forecast_type: One of :data:`nemseer.forecast_types`.
    Returns:
        DataFrame with appropriate datetime filtering applied.
    """
    (runtime_col, forecasted_col) = (
        RUNTIME_COL[forecast_type],
        FORECASTED_COL[forecast_type],
    )
    for (start, end), col in zip(
        ((run_start, run_end), (forecasted_start, forecasted_end)),
        (runtime_col, forecasted_col),
    ):
        if col in df.columns:
            df = _filter_on_datetime_col(df, col, start, end)
    return df


def to_xarray(df: pd.DataFrame, forecast_type: str):
    """Converts a :class:`pandas.DataFrame` to a :class:`xarray.Dataset` using nemseer
    definitions to determine Dataset dimensions.

    The conversion is processed in chunks. If system memory usage exceeds 95%, the
    conversion is terminated with a MemoryError. This is more informative than
    the system killing the Python process.

    Args:
        df: pandas.DataFrame to be converted.
        forecast_type: One of :data:`nemseer.forecast_types`.
    Returns:
        :class:`xarray.Dataset`.
    Warning:
        Raises a warning when attempting to convert high-dimensional data.
    """

    def _determine_multiindex(
        df: pd.DataFrame, forecast_type: str
    ) -> Tuple[pd.MultiIndex, List[str]]:
        """Ascertains which DataFrame columns are in nemseer's datetime, ID or
        type column lists and then uses these columns to create a MultiIndex.

        Args:
            df: pandas DataFrame.
            forecast_type: One of :data:`nemseer.forecast_types`.
        Returns:
            A MultiIndex that will be parsed as dimensions when converting to
            :class:`xarray.Dataset`, and a list of the columns that the MultiIndex
            incorporates.
        """
        multiindex_cols = []
        names = []
        for name, col in zip(
            ("run_time", "forecasted_time"),
            (RUNTIME_COL[forecast_type], FORECASTED_COL[forecast_type]),
        ):
            if col in df.columns:
                names.append(name)
                multiindex_cols.append(col)
        for col in [col for col in df.columns if col not in multiindex_cols]:
            if col in ID_COLS or col in TYPE_COLS:
                names.append(col)
                multiindex_cols.append(col)
        multiindex = pd.MultiIndex.from_frame(df[multiindex_cols], names=names)
        return multiindex, multiindex_cols

    def _df_to_xarray(df: pd.DataFrame) -> xr.Dataset:
        """Reformats supplied DataFrame to a MultiIndexed DataFrame and then
        converts to :class:`xarray.Dataset`.

        Args:
            df: pandas DataFrame.
        Returns:
            DataFrame converted to xarray Dataset.
        """
        multiindex, multiindex_cols = _determine_multiindex(df, forecast_type)
        df = df.set_index(multiindex)
        df = df.drop(multiindex_cols, axis=1)
        ds = df.to_xarray()
        return ds  # type: ignore

    dim_cols = [
        col
        for col in df.columns
        if (
            col in ID_COLS
            or col in TYPE_COLS
            or col in RUNTIME_COL[forecast_type]
            or col in FORECASTED_COL[forecast_type]
        )
    ]
    if len(dim_cols) >= 5 or any(
        [col for col in df.columns if len(df[col].unique()) > 300]
    ):
        logger.warning(
            "High-dimensional data. Large datetime requests may result in the Python "
            + "process being killed by the system"
        )
    if "INTERVENTION" in df.columns and any(df["INTERVENTION"] > 0):
        logger.warning(
            "Intervention periods detected. Discarding intervention runs for conversion"
            + " to xarray. For all data including intervention runs, select conversion"
            + " to pandas DataFrame."
        )
        df = df[df["INTERVENTION"] == 0]
    ds = _df_to_xarray(df)
    return ds



import ast
import logging
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import pyarrow.parquet as pq  # type: ignore
import xarray as xr
from attrs import converters, define, field, validators
from dateutil.relativedelta import relativedelta

from .data import DATETIME_FORMAT, ENUMERATED_TABLES, FORECAST_TYPES

logger = logging.getLogger(__name__)


def _dt_converter(value: str) -> datetime:
    """Convert string to datetime.

    Args:
        value: String with format %Y/%m/%d %H:%M
    Returns:
        Datetime object
    Raises:
        ValueError: If provided datetime string is invalid
    """
    try:
        dt = datetime.strptime(value, DATETIME_FORMAT)
        return dt
    except ValueError:
        try:
            dt = datetime.strptime(value, DATETIME_FORMAT + ":%S")
            if dt.second != 0:
                raise ValueError("If seconds provided in datetime, must be zero.")
            else:
                return dt
        except ValueError:
            raise ValueError(
                "Datetime invalid."
                + " Datetime should be provided as yyyy/mm/dd HH:MM, "
                + " or yyyy/mm/dd HH:MM:00."
            )


def _tablestr_converter(value: Union[str, List[str]]) -> List[str]:
    """Returns a list of table strings, even if a single string is provided

    Args:
        value: Table string or list of table strings
    Returns:
        List of strings
    """
    if type(value) is str:
        return [value]
    else:
        return list(value)


def _validate_run_chronology(instance, attribute, value):
    """Validates run_start against run_end"""
    if value > instance.run_end:
        raise ValueError(
            "Forecast end datetime must be greater than or equal to"
            + " run start datetime."
        )


def _validate_forecasted_chronology(instance, attribute, value):
    """Validated forecasted_start against forecasted_end"""
    if value > instance.forecasted_end:
        raise ValueError(
            "Forecasted end datetime must be greater than or equal to"
            + " forecasted start datetime."
        )


def _validate_relative_chronology(instance, attribute, value) -> None:
    """Validates run_start against forecasted_start"""
    if value > instance.forecasted_start:
        raise ValueError(
            "Forecasted start datetime should be at or after run start datetime."
        )


def _validate_path(instance, attribute, value) -> None:
    """Check the path is a directory and creates it if it is not"""
    if not value.is_dir():
        value.mkdir()
        logger.info(f"Created directory at {value.absolute()}")


def _validate_raw_not_processed(instance, attribute, value) -> None:
    """Check that :attr:`raw_cache` and :attr:`processed_cache` are distinct."""
    if instance.processed_cache:
        if value.absolute() == instance.processed_cache.absolute():
            raise ValueError(
                f"{attribute.name} should be distinct from processed_cache"
            )


def _enumerate_tables(tables: List[str], table_str: str, range_to: int) -> List[str]:
    """Given a table name, populates a list with enumerated table names

    For example, given 'CONSTRAINTSOLUTION' and `range_to`=3, will populate
    `tables` with ['CONSTRAINTSOLUTION1',...,'CONSTRAINTSOLUTION3'].

    Args:
        tables: Table list
        table_str: Table string to enumerate
        range_to: Integer to enumerate to
    Returns:
        `tables` with enumerated `table_str`
    """
    tables.remove(table_str)
    for i in range(1, range_to + 1):
        tables.append(f"{table_str}{i}")
    return tables


def _construct_sqlloader_filename(
    year: int, month: int, forecast_type: str, table: str
) -> str:
    """Constructs filename without file type

    Args:
        year: Year
        month: Month
        forecast_type: One of :data:`nemseer.forecast_types`. See :term:`forecast types`
        table: The name of the table required
    Returns:
        Filename string without file type
    """
    (stryear, strmonth) = (str(year), str(month).rjust(2, "0"))
    if forecast_type == "PREDISPATCH" and table != "MNSPBIDTRK":
        prefix = f"PUBLIC_DVD_{forecast_type}{table}"
    else:
        prefix = f"PUBLIC_DVD_{forecast_type}_{table}"
    fn = prefix + f"_{stryear}{strmonth}010000"
    return fn


def generate_sqlloader_filenames(
    run_start: datetime,
    run_end: datetime,
    forecast_type: str,
    tables: List[str],
) -> Dict[Tuple[int, int, str], str]:
    """Generates MMSDM Historical Data SQLLoader file names based on provided query data

    Returns a tuple of query metadata (`table`, `year`, `month`) mapped to each filename

    Args:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecast_type: One of :data:`nemseer.forecast_types`.
        tables: Table or tables required, provided as a List.
    Returns:
        A tuple of query metadata (`table`, `year`, `month`) mapped to each
        format-agnostic (:term:`SQLLoader`) filename
    """

    def _determine_delta_months(start: datetime, end: datetime):
        """Determines the widest range of months that encompass :attr:`start` and
        :attr:`end.

        Edge cases must be appropriately handled:
            - 2014/05/31 and 2014/06/01 have relativedelta of 1 day, but two
              data months (05/2014 and 06/2014) are required.
            - 2014/05/31 23:00 to 2014/06/01 00:00 only require data for 05/2014

        Args:
            start: datetime of start
            end: datetime of end
        Returns
            delta_months, the total number of months to consider
        """
        MONTH = relativedelta(months=1)
        full_delta = relativedelta(end, start)
        delta_months = full_delta.months * MONTH
        if full_delta.years > 0:
            delta_months += full_delta.years * MONTH * 12
        if (
            (start + delta_months).month != end.month
            and not (end.day == 1 and end.hour == 0 and end.minute == 0)
            and (
                full_delta.days != 0 or full_delta.hours != 0 or full_delta.minutes != 0
            )
        ):
            delta_months += MONTH
        return delta_months.months + (delta_months.years * 12)

    MONTH = relativedelta(months=1)
    int_months = _determine_delta_months(run_start, run_end)
    intervening_dates = [run_start + x * MONTH for x in range(0, int_months + 1)]
    filename_data = {}
    for ftype in ENUMERATED_TABLES:
        if forecast_type == ftype:
            for table, enumerate_to in ENUMERATED_TABLES[ftype]:
                if table in tables:
                    tables = _enumerate_tables(tables, table, enumerate_to)
    for table in tables:
        for date in intervening_dates:
            (year, month) = (date.year, date.month)
            fname = _construct_sqlloader_filename(year, month, forecast_type, table)
            filename_data[(year, month, table)] = fname
    return filename_data


@define
class Query:
    """:class:`Query` validates user inputs and dispatches data downloaders and
    compilers

    Construct :class:`Query` using the :meth:`Query.initialise()` constructor. This
    ensures query metadata is constructed approriately.

    Query:

    - Validates user input data
        - Checks datetimes fit :attr:`yyyy/mm/dd HH:MM` format
        - Checks datetime chronology (e.g. end is after start)
        - Checks requested datetimes are valid for each :term:`forecast type`
        - Validates :term:`forecast type`
        - Validates user-requested tables against what is available on NEMWeb
    - Retains query metadata (via constructor class method
      :meth:`nemseer.query.Query.initialise`)
    - Can check :attr:`raw_cache` and :attr:`processed_cache` contents to streamline
      query compilation

    Attributes:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
        forecast_type: One of :data:`nemseer.forecast_types`.
        tables: Table or tables required. A single table can be supplied as
            a string. Multiple tables can be supplied as a list of strings.
        metadata: Metadata dictionary. Constructed by :meth:`Query.initialise()`.
        raw_cache: Path to build or reuse :term:`raw_cache`.
        processed_cache (optional): Path to build or reuse :term:`processed_cache`.
            Should be distinct from :attr:`raw_cache`
        processed_queries: Defaults to `None` on initialisation. Populated once
            :meth:`Query.find_table_queries_in_processed_cache` is called.

    """

    run_start: datetime = field(
        converter=_dt_converter,
        validator=[_validate_run_chronology, _validate_relative_chronology],
    )
    run_end: datetime = field(converter=_dt_converter)
    forecasted_start: datetime = field(
        converter=_dt_converter, validator=[_validate_forecasted_chronology]
    )
    forecasted_end: datetime = field(converter=_dt_converter)
    forecast_type: str = field(validator=validators.in_(FORECAST_TYPES))
    tables: List[str] = field(converter=_tablestr_converter)
    metadata: Dict[str, str]
    raw_cache: Path = field(
        converter=Path,
        validator=[
            _validate_path,
            _validate_raw_not_processed,
        ],
    )
    processed_cache: Optional[Path] = field(
        default=None,
        converter=converters.optional(Path),
        validator=validators.optional(_validate_path),
    )
    processed_queries: Union[Dict[str, Path], Dict] = field(default=None)

    @classmethod
    def initialise(
        cls,
        run_start: str,
        run_end: str,
        forecasted_start: str,
        forecasted_end: str,
        forecast_type: str,
        tables: Union[str, List[str]],
        raw_cache: str,
        processed_cache: Optional[str] = None,
    ) -> "Query":
        """Constructor method for :class:`Query`. Assembles query metatdata."""
        metadata = {
            "run_start": run_start,
            "run_end": run_end,
            "forecasted_start": forecasted_start,
            "forecasted_end": forecasted_end,
            "forecast_type": forecast_type,
        }
        return cls(
            run_start=run_start,  # type: ignore
            run_end=run_end,  # type: ignore
            forecasted_start=forecasted_start,  # type: ignore
            forecasted_end=forecasted_end,  # type: ignore
            forecast_type=forecast_type,  # type: ignore
            tables=tables,  # type: ignore
            metadata=metadata,  # type: ignore
            raw_cache=raw_cache,  # type: ignore
            processed_cache=processed_cache,  # type: ignore
        )

    def check_all_raw_data_in_cache(self) -> bool:
        """Checks whether *all* requested data is already in the :attr:`raw_cache` as
        parquet

        :meth:`nemseer.downloader.ForecastTypeDownloader.download_csv()`
        handles partial :attr:`raw_cache` completeness

        If all requested data is already in the :attr:`raw_cache` as parquet,
        returns True. Otherwise returns False.
        """
        fnames = generate_sqlloader_filenames(
            self.run_start, self.run_end, self.forecast_type, self.tables
        ).values()
        check = [
            (self.raw_cache / Path(fname + ".parquet")).exists() for fname in fnames
        ]
        if all(check):
            logger.info(f"Query raw data already downloaded to {self.raw_cache}")
            return True
        else:
            return False

    def find_table_queries_in_processed_cache(self, data_format: str) -> None:
        """Determines which tables already have queries saved in the
        :attr:`processed_cache`.

        If data_format=df, this function will sieve through the metadata of all parquet
        files in the :attr:`processed_cache`. Note that parquet metadata is UTF-8
        encoded. Similarly, data_format=xr will check the metadata of all netCDF files.

        Modifies :attr:`Query.processed_queries` from :class:`None` to a :class:`dict`.

        The :class:`dict` is empty if:

        1. :attr:`processed_cache` is :class:`None`
        2. No portion of the query has been saved in the :attr:`processed_cache`

        If a portion of the queries are saved in the :attr:`processed_cache`, then
        :attr:`Query.processed_queries` will be equal to a :class:`dict` that maps
        the saved query's table name to the saved query's filename.

        Args:
            data_format: As per :func:`nemseer.compile_data`
        """
        tables_in_pcache: Union[Dict[str, Path], Dict] = {}
        if not self.processed_cache:
            pass
        else:
            if data_format == "df":
                for file in self.processed_cache.glob("*.parquet"):
                    byte_metadata = pq.read_metadata(file).metadata
                    metadata = ast.literal_eval(
                        (byte_metadata["nemseer".encode()]).decode()
                    )
                    if (
                        metadata_table := metadata.pop("table")
                    ) in self.tables and metadata == self.metadata:
                        tables_in_pcache[metadata_table] = file
                    else:
                        continue
            elif data_format == "xr":
                for file in self.processed_cache.glob("*.nc"):
                    metadata = xr.open_dataset(file).attrs
                    if (
                        metadata_table := metadata.pop("table")
                    ) in self.tables and metadata == self.metadata:
                        tables_in_pcache[metadata_table] = file
                    else:
                        continue
            self.processed_queries = tables_in_pcache



import logging
import re
from datetime import datetime
from pathlib import Path
from typing import Callable, Dict, List, Union

import pandas as pd
import pyarrow as pa  # type: ignore
import pyarrow.parquet as pq  # type: ignore
import xarray as xr
from attrs import define, field

from .data import ENUMERATED_TABLES, INVALID_STUBS_FILE
from .data_handlers import apply_run_and_forecasted_time_filters, to_xarray
from .forecast_type.validators import (
    validate_MTPASA_datetime_inputs,
    validate_P5MIN_datetime_inputs,
    validate_PDPASA_datetime_inputs,
    validate_PREDISPATCH_datetime_inputs,
    validate_STPASA_datetime_inputs,
)
from .query import Query, _enumerate_tables, generate_sqlloader_filenames

logger = logging.getLogger(__name__)


def _map_files_to_table(
    run_start: datetime,
    run_end: datetime,
    forecast_type: str,
    tables: List[str],
) -> Dict[str, List[str]]:
    """Maps filenames of interest to each queried table

    Translates output from `generate_sqlloader_filenames` to map filenames of interest
    to a queried table name.

    Handles enumerated tables (e.g. `PREDISP_ALL_DATA`) by mapping enumerated filenames
    to non-enumerated table type (which is used in the user query).
    E.g. '...PREDISPATCHLOAD1' and '...PREDISPATCHLOAD2' mapped to 'LOAD'

    Args:
        forecast_start: Forecasts made at or after this datetime are queried.
        forecast_end: Forecasts made before or at this datetime are queried.
        forecast_type: One of :data:`nemseer.forecast_types`
        tables: Tables queried.
    Returns:
        A dictionary mapping the queried table name to filenames associated with that
        queried table.
    """
    metadata_to_filename = generate_sqlloader_filenames(
        run_start, run_end, forecast_type, tables
    )
    if forecast_type in ENUMERATED_TABLES.keys():
        enumerated_tables = [pair[0] for pair in ENUMERATED_TABLES[forecast_type]]
    else:
        enumerated_tables = []
    table_file_map: Dict[str, List[str]] = {}
    for table in tables:
        filenames_to_map = list()
        for metadata in metadata_to_filename.keys():
            if metadata[2] == table:
                filenames_to_map.append(metadata_to_filename[metadata])
        if (enum_base := re.match(r"([A-Z]*)[0-9]", table)) and enum_base.group(
            1
        ) in enumerated_tables:
            map_table_name = enum_base.group(1)
            if map_table_name in table_file_map.keys():
                table_file_map[map_table_name].extend(filenames_to_map)
            else:
                table_file_map[map_table_name] = filenames_to_map
        else:
            table_file_map[table] = filenames_to_map
    return table_file_map


def _input_datetime_validation(instance, attribute, value) -> None:
    """Dispatches the correct datetime validator based on the :term:`forecast_type`"""
    validator_map = {
        "P5MIN": validate_P5MIN_datetime_inputs,
        "PREDISPATCH": validate_PREDISPATCH_datetime_inputs,
        "PDPASA": validate_PDPASA_datetime_inputs,
        "STPASA": validate_STPASA_datetime_inputs,
        "MTPASA": validate_MTPASA_datetime_inputs,
    }
    validator_func = validator_map[instance.forecast_type]
    validator_func(
        instance.run_start,
        instance.run_end,
        instance.forecasted_start,
        instance.forecasted_end,
    )
    return None


@define
class DataCompiler:
    """:class:`DataCompiler` compiles data from the :term:`raw_cache` or
    :term:`processed_cache`.

    Attributes:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
        forecast_type: One of :data:`nemseer.forecast_types`.
        tables: Table or tables required. A single table can be supplied as
            a string. Multiple tables can be supplied as a list of strings.
        metadata: Metadata dictionary. Constructed by
            :meth:`Query.initialise() <nemseer.query.Query.initialise()>`.
        raw_cache (optional): Path to build or reuse :term:`raw_cache`.
        processed_cache (optional): Path to build or reuse :term`processed cache`.
            Should be distinct from :attr:`raw_cache`
        processed_queries: Defaults to :class:`None` on initialisation.
        raw_table: Populated via :meth:`DataCompiler.from_Query`
        compiled_data: Defaults to :class:`None` on initialisation. Populated once data
            is compiled by methods.
    """

    run_start: datetime = field(validator=_input_datetime_validation)
    run_end: datetime
    forecasted_start: datetime
    forecasted_end: datetime
    forecast_type: str
    metadata: Dict[str, str]
    raw_cache: Path
    processed_cache: Union[None, Path]
    processed_queries: Union[Dict[str, Path], Dict]
    raw_tables: List[str]
    compiled_data: Union[None, Dict[str, pd.DataFrame], Dict[str, xr.Dataset]] = field(
        default=None
    )

    @classmethod
    def from_Query(cls, query: Query) -> "DataCompiler":
        """Constructor method for :class:`DataCompiler` from
        :class:`Query <nemseer.query.Query>`."""
        tables = query.tables
        if query.processed_cache:
            if query.processed_queries:
                raw_tables = list(set(tables) - set(query.processed_queries.keys()))
            else:
                raw_tables = tables
        else:
            raw_tables = tables
        for ftype in ENUMERATED_TABLES:
            if query.forecast_type == ftype:
                for table, enumerate_to in ENUMERATED_TABLES[ftype]:
                    if table in raw_tables:
                        tables = _enumerate_tables(tables, table, enumerate_to)
        return cls(
            run_start=query.run_start,
            run_end=query.run_end,
            forecasted_start=query.forecasted_start,
            forecasted_end=query.forecasted_end,
            forecast_type=query.forecast_type,
            metadata=query.metadata,
            raw_cache=query.raw_cache,
            processed_cache=query.processed_cache,
            processed_queries=query.processed_queries,
            raw_tables=raw_tables,
            compiled_data=None,
        )

    def invalid_or_corrupted_files(self) -> List[str]:
        """A list of invalid/corrupted files as per files in `.invalid_aemo_files.txt`.
        Returns an empty list if the stubfile does not exist.
        """
        invalid_or_corrupted_stubfile = self.raw_cache / Path(INVALID_STUBS_FILE)
        if invalid_or_corrupted_stubfile.exists():
            with open(invalid_or_corrupted_stubfile, "r") as f:
                invalid_or_corrupted = f.readlines()
            check_files = [f.strip() for f in invalid_or_corrupted]
            return check_files
        else:
            return []

    def compile_raw_data(self, data_format: str = "df") -> None:
        """Compiles data from :attr:`raw_cache` to a :class:`pandas.DataFrame` (default)
        or to a :class:`xarray.Dataset`.

        This compiler will:

        - Skip invalid/corrupted files as recorded in `.invalid_aemo_files.txt`
        - Read :term:`raw_cache` parquet files and apply datetime filtering
        - Convert :class:`DataFrame <pandas.DataFrame>` to :class:`xarray.Dataset`
          (if :attr:`data_format` = "xr")
        - Update :attr:`compiled_data`

        Args:
            data_format: Default "df" (:class:`pandas.DataFrame`). Other valid input is
                "xr", which returns :class:`xarray.Dataset`.
        Warning:
            Skips any files previously found to be invalid/corrupted and prints a
            warning
        """
        file_to_table_map = _map_files_to_table(
            self.run_start, self.run_end, self.forecast_type, self.raw_tables
        )
        table_to_data_map = {}
        invalid_files = self.invalid_or_corrupted_files()
        for table in file_to_table_map.keys():
            files = file_to_table_map[table]
            filtered_files = [file for file in files if file not in invalid_files]
            if not filtered_files:
                raise ValueError(
                    "Query failed as all files to be compiled were found to be"
                    + " invalid/corrupt on previous download. You can force nemseer"
                    + " to load these files by deleting them from "
                    + ".invalid_aemo_files.txt"
                )
            elif len(filtered_files) < len(files):
                logger.warning(
                    "Some files not compiled as these were found to be"
                    + " invalid/corrupt on previous download. You can force nemseer"
                    + " to load these files by deleting them from "
                    + ".invalid_aemo_files.txt"
                )
            dfs = []
            for file in filtered_files:
                filepath = self.raw_cache / Path(f"{file}.parquet")
                df = pd.read_parquet(filepath)
                df = apply_run_and_forecasted_time_filters(
                    df,
                    self.run_start,
                    self.run_end,
                    self.forecasted_start,
                    self.forecasted_end,
                    self.forecast_type,
                )
                dfs.append(df.reset_index(drop=True))
            concat_df = pd.concat(dfs)
            if any(concat_df.duplicated()):
                logger.warning(
                    "Duplicate rows detected whilst concatenating data. "
                    + "Dropping these rows."
                )
                concat_df = concat_df.drop_duplicates()
            if data_format == "xr":
                logger.info(f"Converting {table} data to xarray.")
                concat_data = to_xarray(concat_df, self.forecast_type)
            else:
                concat_data = concat_df
            table_to_data_map[table] = concat_data

        if not self.compiled_data:
            self.compiled_data = table_to_data_map
        else:
            self.compiled_data.update(table_to_data_map)

    def compile_processed_data(self, data_format: str = "df") -> None:
        """Compiles data from the :attr:`processed_cache`, as per entries in
        :attr:`processed_queries`, to a :class:`pandas.DataFrame` (default) or to a
        :class:`xarray.Dataset`.

        This method will update :attr:`compiled_data`.

        Args:
            data_format: Default "df" (:class:`pandas.DataFrame`). Other valid input
                is "xr", which compiles :class:`xarray.Dataset`.
        """
        read_fn: Dict[str, Callable] = {
            "df": pd.read_parquet,
            "xr": xr.open_dataset,
        }
        processed_data = {}
        if not self.processed_queries:
            pass
        else:
            for table in self.processed_queries:
                file = self.processed_queries[table]
                logger.info(f"Compiling {table} data from the processed cache")
                data = read_fn[data_format](file)
                processed_data[table] = data
            if not self.compiled_data:
                self.compiled_data = processed_data
            else:
                self.compiled_data.update(processed_data)

    def write_to_processed_cache(self) -> None:
        """Writes netCDF4 for :class:`xarray.Dataset` and parquet
        for :class:`pandas.DataFrame` to the :attr:`processed_cache` with associated
        query metadata.

        Note that parquet metadata needs to be UTF-8 encoded.

        Raises:
            ValueError: If :attr:`processed_cache` is :class:`None`, or if
                :attr:`compiled_data` contains data that is neither all
                :class:`pandas.DataFrame` or all :class:`xarray.Dataset`
            IOError: If :attr:`compiled_data` is :class:`None`
        """

        def _df_to_pyarrow_with_metadata(
            df: pd.DataFrame, metadata: Dict[str, str]
        ) -> pa.Table:
            """Converts DataFrame to pyarrow Table so that metadata can be added.
            Args:
                df: pandas DataFrame
                metadata: :class:`dict` built by
                    :classmethod:`nemseer.query.Query.initialise()`
            Returns:
                pyarrow Table with schema and nemseer metadata encoded as a b-string
            """
            table = pa.Table.from_pandas(df)
            pandas_metadata = table.schema.metadata
            nemseer_metadata = {b"nemseer": str(metadata).encode()}
            merged_metadata = {**pandas_metadata, **nemseer_metadata}
            table = table.replace_schema_metadata(merged_metadata)
            return table

        def _build_query_filename(compiler: DataCompiler, table: str) -> str:
            """Builds a filename based on a table name and query details.

            Args:
                compiler: DataCompiler instance with populated query info
                table: Specific table to build a filename for
            Returns
                A filename constructed based on query details.
            """
            (fs, fe) = (compiler.forecasted_start, compiler.forecasted_end)
            (rs, re) = (compiler.run_start, compiler.run_end)
            rs_re = (
                f"{rs.year}{rs.month}{rs.day}{rs.hour}{rs.minute}"
                + "_"
                + f"{re.year}{re.month}{re.day}{re.hour}{re.minute}"
            )
            fs_fe = (
                f"{fs.year}{fs.month}{fs.day}{fs.hour}{fs.minute}"
                + "_"
                + f"{fe.year}{fe.month}{fe.day}{fe.hour}{fe.minute}"
            )
            fn = f"{compiler.forecast_type}_{table}_{rs_re}_{fs_fe}"
            return fn

        if self.processed_cache is None:
            raise ValueError(
                "Writing to processed cache requires that the processed cache "
                + "be specified"
            )
        if self.compiled_data is None:
            raise IOError("No compiled data to write to processed cache")
        data = self.compiled_data
        xrbool = all([type(data) is xr.Dataset for data in self.compiled_data.values()])
        dfbool = all(
            [type(data) is pd.DataFrame for data in self.compiled_data.values()]
        )
        for table in data.keys():
            if self.processed_queries and table in self.processed_queries.keys():
                continue
            else:
                fn = _build_query_filename(self, table)
                self.metadata.update({"table": table})
                dataset = data[table]
                if xrbool:
                    fn_path = self.processed_cache / Path(fn + ".nc")
                    dataset.attrs = self.metadata  # type: ignore
                    logger.info(f"Writing {table} to the processed cache as netCDF")
                    dataset.to_netcdf(fn_path)  # type: ignore
                elif dfbool:
                    fn_path = self.processed_cache / Path(fn + ".parquet")
                    pyarrow_table = _df_to_pyarrow_with_metadata(
                        dataset, self.metadata  # type: ignore
                    )
                    logger.info(f"Writing {table} to the processed cache as parquet")
                    pq.write_table(pyarrow_table, fn_path)
                else:
                    raise ValueError(
                        "Compiled data is not in a valid data structure. "
                        + "Compiled data should be in a pandas DataFrame or "
                        + "xarray Dataset"
                    )



from typing import Dict, List, Tuple, Union

import pandas as pd
import xarray as xr

from .data_compilers import DataCompiler
from .downloader import ForecastTypeDownloader
from .forecast_type.run_time_generators import generate_runtimes
from .query import Query


def _initiate_downloads_from_query(query: Query, keep_csv: bool = False) -> None:
    """Initiates download actions using :class:`nemseer.query.Query`
    Args:
        query: :class:`nemseer.query.Query`
    Returns:
        None
    """
    if query.check_all_raw_data_in_cache():
        pass
    else:
        downloader = ForecastTypeDownloader.from_Query(query)
        downloader.download_csv()
        downloader.convert_to_parquet(keep_csv=keep_csv)
    return None


def download_raw_data(
    forecast_type: str,
    tables: Union[str, List[str]],
    raw_cache: str,
    run_start: Union[str, None] = None,
    run_end: Union[str, None] = None,
    forecasted_start: Union[str, None] = None,
    forecasted_end: Union[str, None] = None,
    keep_csv: bool = False,
) -> None:
    """Downloads raw forecast data from NEMWeb MMSDM Historical Data SQLLoader

    Downloads raw forecast data. Accepts a datetime pair, which can be either of:

    1. :attr:`run_start` and :attr:`run_end`
    2. :attr:`forecasted_start` and :attr:`forecasted_end`

    Examples:
        See :ref:`downloading raw data examples <quick_start:downloading raw data>`.

    Arguments:
        forecast_type: One of :data:`nemseer.forecast_types`
        tables: Table or tables required. A single table can be supplied as
            a string. Multiple tables can be supplied as a list of strings.
        raw_cache: Path to create or reuse as :term:`raw_cache`. Files are downloaded
            to this directory and cached data is maintained in the parquet format.
        run_start: Forecast runs at or after this datetime are queried. If supplied,
            must be included with :attr:`run_end`.
        run_end: Forecast runs before or at this datetime are queried. If supplied,
            must be included with :attr:`run_start`.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained. If supplied, must be included with
            :attr:`forecasted_end`.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retained. If supplied, must be included with
            :attr:`forecasted_start`.
        keep_csv: Default False. If True, downloaded csvs are retained in the
            :term:`raw_cache`.
    Raises:
        ValueError: If a valid pair of datetimes is not supplied, or if more than a
            valid pair of datetimes is supplied.
    """

    def _generate_other_datetime_pair(
        start: str,
        end: str,
        input_datetime_type: str,
        forecast_type: str,
    ) -> Tuple[str, str]:
        """Given forecasted times, generates runtimes and vice versa.

        Args:
            start: Start datetime.
            end: End datetime.
            input_datetime_type: Specified whether :attr:`start` and :attr:`end`
                correspond to 'run' or 'forecasted' datetimes
            forecast_type: One of :data:`nemseer.forecast_types`
        Returns:
            A tuple of datetime strings that correspond to the "other" set of datetimes.
        Raises:
            ValueError: If :attr:`input_datetime_type` does not correspond to 'run'
                or 'forecasted'
        """
        if input_datetime_type not in ("run", "forecasted"):
            raise ValueError("Input datetime type must be 'run' or 'forecasted'")
        if input_datetime_type == "forecasted":
            (other_start, other_end) = generate_runtimes(start, end, forecast_type)
        else:
            other_start = start
            other_end = start
        return other_start, other_end

    if run_start and run_end and not forecasted_start and not forecasted_end:
        forecasted_start, forecasted_end = _generate_other_datetime_pair(
            run_start, run_end, "run", forecast_type
        )
    elif forecasted_start and forecasted_end and not run_start and not run_end:
        run_start, run_end = _generate_other_datetime_pair(
            forecasted_start, forecasted_end, "forecasted", forecast_type
        )
    else:
        raise ValueError(
            "Provide both of run_start and run_end (and no forecasted times),"
            + " or both of forecasted_start and forecasted_end (and no run times)."
        )
    query = Query.initialise(
        run_start=run_start,
        run_end=run_end,
        forecasted_start=forecasted_start,
        forecasted_end=forecasted_end,
        forecast_type=forecast_type,
        tables=tables,
        raw_cache=raw_cache,
    )
    _initiate_downloads_from_query(query, keep_csv=keep_csv)


def compile_data(
    run_start: str,
    run_end: str,
    forecasted_start: str,
    forecasted_end: str,
    forecast_type: str,
    tables: Union[str, List[str]],
    raw_cache: str,
    processed_cache: Union[None, str] = None,
    data_format: str = "df",
) -> Union[Dict[str, pd.DataFrame], Dict[str, xr.Dataset], None]:
    """Compiles queried data from :attr:`raw_cache` and/or :attr:`processed_cache`.

    For each queried table, this function:

    1. If required, downloads raw forecast data for the table and converts to the
       requested data structure.
    2. Otherwise, compiles table data from either of or both of the caches.
    3. Applies user-requested filtering to :term:`run times` and
       :term:`forecasted times` to any raw data.


    If :attr:`data_format` = "df" (default), a :class:`pandas.DataFrame` is returned.
    Otherwise, if :attr:`data_format` = "xr", a :class:`xarray.Dataset` is returned.

    Examples:
        See :ref:`compiling data examples <quick_start:compiling data>`.

    Arguments:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retained.
        forecast_type: One of :data:`nemseer.forecast_types`
        tables: Table or tables required. A single table can be supplied as
            a string. Multiple tables can be supplied as a list of strings.
        raw_cache: Path to create or reuse as :term:`raw_cache`. Files are downloaded
            to this directory and cached data is maintained in the parquet format.
        processed_cache (optional): Path to build or reuse :term:`processed_cache`.
            Should be distinct from :attr:`raw_cache`
        data_format: Default is 'df', which returns :class:`pandas DataFrame`.
            Can also request 'xr', which returns :class:`xarray.Dataset`.
    """
    if data_format not in (fmts := ("df", "xr")):
        raise ValueError(f"Invalid data format. Formats include: {fmts}")
    query = Query.initialise(
        run_start=run_start,
        run_end=run_end,
        forecasted_start=forecasted_start,
        forecasted_end=forecasted_end,
        forecast_type=forecast_type,
        tables=tables,
        raw_cache=raw_cache,
        processed_cache=processed_cache,
    )
    query.find_table_queries_in_processed_cache(data_format=data_format)
    compiler = DataCompiler.from_Query(query)
    if compiler.raw_tables:
        _initiate_downloads_from_query(query, keep_csv=False)
        compiler.compile_raw_data(data_format=data_format)
    compiler.compile_processed_data(data_format=data_format)
    if compiler.processed_cache:
        compiler.write_to_processed_cache()
    data = compiler.compiled_data
    return data
